(defgroup trex-mode nil
  "Trex mode."
  :prefix "trex-"
  :group 'languages)

(defface trex-instruction-keywords-face
  '((t (:height 1.0 :weight bold :foreground "DarkCyan")))
  "Face used to highlight instruction keywords.")

(defface trex-expression-keywords-face
  '((t (:height 1.0 :foreground "LightGreen")))
  "Face used to highlight expression keywords.")

(defface trex-comments-face
  '((t (:height 1.0 :foreground "OrangeRed")))
  "Face used to highlight comments.")

(defface trex-definition-face
  '((t (:height 1.0 :foreground "sienna2")))
  "Face used to highlight definitions.")
;; (set-face-attribute 'trex-definition-face nil
;;                               :foreground "sienna2")

(defface trex-default-face
  '((t (:height 1.0 :foreground "LightBlue")))
  "Face used to highlight normal text.")

(defun make-instruction-keyword (kw)
  (concat "\\([ 	]+\\|^\\)" kw "\\([ 	]+\\|$\\|(\\)"))

(defun make-expression-keyword (kw)
  (concat "\\b" kw "\\b"))

(defvar trex-font-lock-keywords-1
  (list
   (list
    (concat "\\(" "(\\*.*[$.]*\\*)\\)")
    '(1 'trex-comments-face nil))
   (list
    (concat
     "\\(" (make-instruction-keyword "if") "\\|"
     (make-instruction-keyword "then") "\\|"
     (make-instruction-keyword "else") "\\|"
     (make-instruction-keyword "genrule") "\\|"
     (make-instruction-keyword "for") "\\|"
     (make-instruction-keyword "from") "\\|"
     (make-instruction-keyword "to") "\\|"
     (make-instruction-keyword "in") "\\|"
     (make-instruction-keyword "do") "\\|"
     (make-instruction-keyword "setn") "\\|"
     (make-instruction-keyword "set") "\\|"
     (make-instruction-keyword "import") "\\|"
     (make-instruction-keyword "load") "\\|"
     ";\\|,\\|=\\|(\\|)\\|<>\\|<\\|>\\)"
     )
    '(1 'trex-instruction-keywords-face nil))
   (list
    (concat
     "\\(" "\\$[a-zA-Z0-9_]+"
     "\\)"
     )
    '(1 'trex-definition-face nil))
   (list
    (concat "\\(._.\\|_.\\|._"
	    "\\)")
    '(1 'trex-default-face nil))
   (list
    (concat
     "\\(" (make-expression-keyword "add") "\\|"
     (make-expression-keyword "getn") "\\|"
     (make-expression-keyword "subsn") "\\|"
     (make-expression-keyword "op") "\\|"
     (make-expression-keyword "var") "\\|"
     (make-expression-keyword "mem") "\\|"
     (make-expression-keyword "concat") "\\|"
     (make-expression-keyword "append") "\\|"
     (make-expression-keyword "seq") "\\|"
     (make-expression-keyword "useq") "\\|"
     (make-expression-keyword "cseq") "\\|"
     (make-expression-keyword "eval") "\\|"
     (make-expression-keyword "len") "\\|"
     (make-expression-keyword "symboldepth") "\\|"
     (make-expression-keyword "or") "\\|"
     (make-expression-keyword "and") "\\|"
     (make-expression-keyword "not") "\\|"
     "&&" "\\|" "||" "\\)"
     )
    '(1 'trex-expression-keywords-face nil))
   (list
    (concat "\\(.\\)")
    '(1 'trex-default-face nil))
   "Description of regular expressions for highlighting"
   ))

(defconst trex-indent 4
  "This variable contains the indentation in trex-mode.")

(defconst trex-end-indent 4
  "This variable indicates the relative position of the
\"end\" in TREX mode.")

(defun trex-indent-line ()
  (interactive)
  (beginning-of-line)
  (if (bobp) (indent-line-to 0)
    (let ((not-indented t) cur-indent)
      (if (looking-at "[ \t]*)\\(;\\|[ \t]*else\\)[ \t]*$")
	  (progn (save-excursion
		   (forward-line -1)
		   (setq cur-indent (- (current-indentation) trex-end-indent))
		   )
		 (if (< cur-indent 0) (setq cur-indent 0)
		  ))
	(save-excursion
	  (while not-indented
	    (forward-line -1)
	    (if (looking-at "[ \t]*)\\(;\\|[ \t]*else\\)[ \t]*$")
		(progn (setq cur-indent (current-indentation))
		       (setq not-indented nil))
	      (if (looking-at "^.*([ \t]*$")
		  (progn (setq cur-indent (+ (current-indentation) trex-indent))
			 (setq not-indented nil))
		(if (bobp) (setq not-indented nil))
		)
	      )
	    )
	  )
	)
      (if cur-indent (indent-line-to cur-indent)
	(indent-line-to 0))
      )
    )
  )

(defun trex-mode ()
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'trex-mode)
  (setq mode-name "trex-mode")
  "Major mode for editing trex code."
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(trex-font-lock-keywords-1))
  (make-local-variable 'comment-start)
  (setq comment-start "(* ")
  (make-local-variable 'comment-end)
  (setq comment-end " *)")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "(\\*+[ \t]*")
  (make-local-variable 'comment-multi-line)
  (setq comment-multi-line t)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'trex-indent-line)
)

(provide 'trex)