import java_bytecode;

(* Definitions *)
set locals = seq(local,$max_locals);
set local_vars = seq(var(l),$max_locals);
set field_vars = seq(var(f),$nb_fields);
set heap_vars = seq(var(h),$nb_classes);
set pcs = seq(pp,$max_pc);

(* Creation of objects constructors of type object0(var(f0),...), ... *)
(* and object0(nilfield,...), ... *)
(* and objectO(zero,...), object1(val_null,...), ... representing objects *)
(* initialized by Java.lang.Object constructor *)
set objects_vars = cseq(val_null,$nb_classes);
set uninitialized_objects = cseq(val_null,$nb_classes);
set initialized_objects = cseq(val_null,$nb_classes);

let f_used = 0 in (
  for i from 1 to $nb_classes
  do(
    let nb_fields_i = getn($nb_fields_per_classes,$i) in (
      let field_vars_i = seq(var(f),$nb_fields_i)
      and nilfields_i = cseq(nilfield,$nb_fields_i)
      and init_fields_i = getn($init_fields_per_classes,$i) in (
	for j from 1 to $nb_fields_i
	do(
	  set f_used = add(1,$f_used);
	  let f_var = getn($field_vars,$f_used) in (
	    setn(field_vars_i,$j) = $f_var;
	  );
	);
	let o = add($i,-1) in (
	  setn(objects_vars,$i) = concat(object,($o,$field_vars_i));
	  setn(uninitialized_objects,$i) = concat(object,($o,$nilfields_i));
	  setn(initialized_objects,$i) = concat(object,($o,$init_fields_i));
	);
      );
    );
  );
);

defun get_object_var(i) =
(
  getn($objects_vars,$i);
);

defun get_uninitialized_object(i) =
(
  getn($uninitialized_objects,$i);
);

defun get_initialized_object(i) =
(
  getn($initialized_objects,$i);
);

(* Creation of nilstacks constructors of type nilstack0 ... *)
set nilstacks = seq(nilstack,$nb_classes);
defun get_nilstack(i) =
(
  getn($nilstacks,$i);
);

(* Creation of nillocal sequence *)
set nillocals = locals(cseq(nillocal,$max_locals));

(* Creation of object stacks constructors without parameters of type *)
(* stack0(ni, ... *)
set stacks = useq(stack,$nb_classes);
defun get_stack(i) =
(
  getn($stacks,$i);
);

(* alias for main class *)
set main_class = getn($classes,1);

(* alias for initialization method *)
set init_method = Method(<<init>>,NilType,void);

(* alias for stubs *)
set java_lang_string = Class(ConsName(<java>,ConsName(<lang>,ConsName(<String>,NilName))));
set java_lang_stringbuilder = Class(ConsName(<java>,ConsName(<lang>,ConsName(<StringBuilder>,NilName))));
set java_lang_object = Class(ConsName(<java>,ConsName(<lang>,ConsName(<Object>,NilName))));
set java_lang_system = Class(ConsName(<java>,ConsName(<lang>,ConsName(<System>,NilName))));
set java_io_exception = Class(ConsName(<java>,ConsName(<io>,ConsName(<Exception>,NilName))));
set java_io_inputstream = Class(ConsName(<java>,ConsName(<io>,ConsName(<InputStream>,NilName))));
set java_io_printstream = Class(ConsName(<java>,ConsName(<io>,ConsName(<PrintStream>,NilName))));

(* definition of initialJavaState *)
set iframe = frame(name(Method(<main>,ConsType(AType(
OType($java_lang_string)),NilType),void),$main_class),pp0,nilstack,$nillocals);

set iheap_args = $nilstacks;
for i from 1 to $nb_classes
do(
  setn(iheap_args,$i) = heap(zero,getn($nilstacks,$i));
);

let input_class_num = mem($java_io_inputstream,$classes)
and print_class_num = mem($java_io_printstream,$classes) in (
  let input_obj = getn($initialized_objects,$input_class_num)
  and print_obj = getn($initialized_objects,$print_class_num) in (
    let input_num = add($input_class_num,-1)
    and print_num = add($print_class_num,-1) in (
      let input_stack = concat(stack,($input_num,($input_obj,getn($nilstacks,$input_class_num))))
      and print_stack = concat(stack,($print_num,($print_obj,getn($nilstacks,$print_class_num)))) in (
	set iheap = heaps($iheap_args);
	set iheap = subsn($iheap,$input_class_num,heap(succ(zero),$input_stack));
	set iheap = subsn($iheap,$print_class_num,heap(succ(zero),$print_stack));
      );
    );
  );
);

let loc_sys_in = loc($java_io_inputstream,zero)
and loc_sys_out = loc($java_io_printstream,zero)
and sys_class_num = mem($java_lang_system,$classes) in (
  let sys_obj = concat(object,(add($sys_class_num,-1),($loc_sys_in,$loc_sys_out))) in (
    set sheap = heaps($initialized_objects);
    set sheap = subsn($sheap,$sys_class_num,$sys_obj);
  );
);

let istate = state($iframe,nilcallstack,$iheap,$sheap) in (
  set iostate = IO($istate,var(x),niloutstack);
);

genrule(initialJavaState(var(x)),$iostate);

genrule(instackExample,instack(succ(succ(succ(zero))),instack(succ(zero),instack(succ(succ(zero)),
instack(zero,instack(pred(zero),nilinstack))))));

defun get_class(i) =
(
  getn($classes,$i);
);

defun get_field(i) =
(
  getn($fields,$i);
);

defun get_class_method_flags(i,j) =
(
  getn(getn($methods_flags_per_classes,$i),$j);
);

defun gen_heaps(i,a,b) =
(
  heaps(subsn($heap_vars,$i,heap($a,$b)));
);

(* Definition of initialization for Java.Lang.Object *)

genrule(
  state(frame(name($init_method,$java_lang_object),pp0,nilstack,var(ll)),
  callstack(storedframe(var(m),var(pc),var(s),var(l)),var(f)),var(h),var(k)),
  state(frame(var(m),next(var(pc)),var(s),var(l)),var(f),var(h),var(k))
);
