open JBasics
open JClass
open Utils
open Terms_of_integer
open Terms_of_string

let consname = "ConsName"
let nilname = "NilName"
let constype = "ConsType"
let niltype = "NilType"
let voidtype = "void"
let class_constructor = "Class"
let field_constructor = "Field"
let method_constructor = "Method"
let array_type = "AType"
let object_type = "OType"
  
let assoc_ts = [
  (`Byte,"TByte");
  (`Char,"TChar");
  (`Double,"TDouble");
  (`Float,"TFloat");
  (`Int,"TInt");
  (`Long,"TLong");
  (`Short,"TShort");
  (`Bool,"TBool")]

let assoc_tinit = [
  (`Byte,"zero");
  (`Char,nilchar);
  (`Double,"zero");
  (`Float,"zero");
  (`Int,"zero");
  (`Long,"zero");
  (`Short,"zero");
  (`Bool,"val_false")]
  
let list2s s n l =
  let len = List.length l in
    match len with
      | 0 -> n
      | _ -> 
	  s ^ "(" ^ (String.concat ("," ^ s ^"(") l)
	  ^ "," ^ n ^ (String.make len ')')

let cn2s (cn : class_name) =
  match cn with
    | [] -> failwith "invalid class name (empty)"
    | l ->
	class_constructor
	^ encaps (list2s consname nilname
		    (List.map (fun x -> encaps_prog x) l))
	  
let rec arg2s = function
  | TBasic b -> List.assoc b assoc_ts
  | TObject o ->
      (match o with
	 | TClass cn -> object_type^encaps(cn2s cn)
	 | TArray a -> array_type^encaps(arg2s a)
      )

let type2init = function
  | TBasic b -> List.assoc b assoc_tinit
  | TObject _ -> "val_null"
	
let fn2s (cn : class_name) (fs : field_signature) =
  field_constructor
  ^ encaps(String.concat ","
	     [cn2s cn; encaps_prog fs.fs_name; arg2s fs.fs_type])

let mn2s (ms : method_signature) =
  let ret2s = function
    | None -> voidtype
    | Some t -> arg2s t
  in
    method_constructor ^ encaps(
      String.concat ","
	[ encaps_prog ms.ms_name;
	  (list2s constype niltype (List.map arg2s ms.ms_parameters));
	  ret2s ms.ms_return_type ])
